import reals._real
import reals._homographic
import reals._computation

from fractions import Fraction
from typing import Optional, Union


class Approximation:
    def __init__(self, x: Union[reals._real.Real, reals._computation.Computation]):
        self.ingestions = 0
        self.state = reals._homographic.Homographic(1, 0, 0, 1)
        if isinstance(x, reals._real.Real):
            self.computation = x.compute()
        else:
            self.computation = x

    def improve(self, n: int = 1) -> None:
        for _ in range(0, n):
            try:
                self.ingestions += 1
                next_term = next(self.computation)
                self.state.ingest(next_term)
            except StopIteration:
                self.state.ingest_inf()
                break

    def improve_epsilon(self, epsilon: Fraction) -> None:
        while not (eps := self.epsilon_fraction()) or eps > epsilon:
            try:
                self.ingestions += 1
                next_term = next(self.computation)
                self.state.ingest(next_term)
            except StopIteration:
                self.state.ingest_inf()
                assert self.epsilon_fraction() == 0
                break

    def as_fraction(self) -> Optional[Fraction]:
        if self.state.c != 0:
            return Fraction(self.state.a, self.state.c)
        return None

    def as_float(self) -> Optional[float]:
        if self.state.c != 0:
            return self.state.a / self.state.c
        return None

    def _lower(self) -> tuple[int, int]:
        if self.ingestions % 2 == 1:
            return (self.state.a, self.state.c)
        else:
            return (self.state.a + self.state.b, self.state.c + self.state.d)

    def _upper(self) -> tuple[int, int]:
        if self.ingestions % 2 == 1:
            return (self.state.a + self.state.b, self.state.c + self.state.d)
        else:
            return (self.state.a, self.state.c)

    def lower_bound_fraction(self) -> Optional[Fraction]:
        if self.ingestions == 0:
            return None
        p, q = self._lower()
        return Fraction(p, q) if q != 0 else None

    def lower_bound_float(self) -> Optional[float]:
        if self.ingestions == 0:
            return None
        p, q = self._lower()
        return p / q if q != 0 else None

    def upper_bound_fraction(self) -> Optional[Fraction]:
        p, q = self._upper()
        return Fraction(p, q) if q != 0 else None

    def upper_bound_float(self) -> Optional[float]:
        p, q = self._upper()
        return p / q if q != 0 else None

    def interval_fraction(self) -> tuple[Optional[Fraction], Optional[Fraction]]:
        return (self.lower_bound_fraction(), self.upper_bound_fraction())

    def interval_float(self) -> tuple[Optional[float], Optional[float]]:
        return (self.lower_bound_float(), self.upper_bound_float())

    def epsilon_fraction(self) -> Optional[Fraction]:
        lower, upper = self.interval_fraction()
        if lower and upper:
            return upper - lower
        return None

    def float_epsilon(self) -> Optional[float]:
        rational_epsilon = self.epsilon_fraction()
        if rational_epsilon:
            return float(rational_epsilon)
        return None

    def closest_float(self) -> float:
        while not (lo := self.lower_bound_float()) or not (hi := self.upper_bound_float()) or lo != hi:
            self.improve()

        assert lo
        return lo


def closest_float(x: reals._real.Real) -> float:
    return Approximation(x).closest_float()


def best_rational_approximations(x: reals._real.Real, n: int):
    a = Approximation(x)

    result: list[Fraction] = []
    for _ in range(0, n):
        a.improve()
        rational_approximation = a.as_fraction()
        assert rational_approximation is not None
        result.append(rational_approximation)

    return result



# TODO(Ruben): this needs to be rewritten to use absolute imports
from reals._real import Real, Number, ensure_real
from reals._normalize import normalize
from reals._computation import Computation
from reals.approximation import Approximation
from reals._algebraic_computation import AlgebraicComputation

from math import factorial
from typing import Generator
from fractions import Fraction


def exp_int_computation(n: int) -> Generator[tuple[int, int], None, None]:
    m = 6
    while True:
        yield (m, n * n)
        m += 4


def exp_int(n: int) -> Real:
    n_squared = n * n
    computation = Real(exp_int_computation(n)).compute()
    return normalize(Real(AlgebraicComputation(computation, (2 + n, n_squared, 2 - n, n_squared))))


# compute exp(x) for small negative x
def exp_small(x: Real) -> Real:
    return Real(ExponentialComputation(x))


def pow(x: Real, n: int) -> Real:
    assert n >= 0
    if n == 0:
        return Real.from_int(1)
    result = x
    for _ in range(0, n - 1):
        result *= x
    return result


# compute exp(x) for small negative x
# Technically this method will converge for any x, the negativeness is needed to make the Taylor series
# alternating so that it can be used to obtain both lower and upper bounds. The smallness of x is needed
# to improve the convergence.
class ExponentialComputation(Computation):
    def __init__(self, x: Real) -> None:
        self.x = x
        self.lo = 1 + x + (x * x) / 2 + (x * x * x) / 6
        self.hi = self.lo + (x * x * x * x) / 24
        self.lo_comp = self.lo.compute()
        self.hi_comp = self.hi.compute()
        self.n = 0
        self.k = 5

    def __next__(self) -> tuple[int, int]:
        while True:
            n_lo = next(self.lo_comp)
            n_hi = next(self.hi_comp)

            if n_lo == n_hi:
                self.n += 1
                return n_lo

            self.lo = self.hi + pow(self.x, self.k) / factorial(self.k)
            self.k += 1
            self.hi = self.lo + pow(self.x, self.k) / factorial(self.k)
            self.k += 1

            self.lo_comp = self.lo.compute()
            self.hi_comp = self.hi.compute()
            for _ in range(0, self.n):
                next(self.lo_comp)
                next(self.hi_comp)


# note that this might not be ideal to calculate exp for x of type Real which turns out to be integer
# (e.g. x = sqrt(3) * sqrt(3))
def exp(x: Number) -> Real:
    # if x is an integer it can be computed more efficiently
    if isinstance(x, int):
        return exp_int(x)

    x = ensure_real(x)

    # get numbers n and rest such that x = n + rest, n is an integer, and rest is a small negative number
    # more specifically, calculate n = ceil(upper_bound(x)) and rest = x - n
    epsilon = Fraction(1, 1000)
    a = Approximation(x)
    a.improve_epsilon(epsilon)
    upper_bound = a.upper_bound_fraction()
    assert upper_bound
    p, q = upper_bound.as_integer_ratio()

    n = (p + q - 1) // q
    rest = x - n

    return exp_int(n) * exp_small(rest)



import reals._real
import reals._computation
import reals._quadratic_computation
import reals._homographic
import reals._term

from fractions import Fraction
from typing import Union


# compute sqrt(n) where n is an integer
class SqrtComputation(reals._computation.Computation):
    def __init__(self, n: int):
        if n < 0:
            raise ValueError()

        self.state = reals._homographic.Homographic(0, n, 1, 0)

    def __next__(self) -> reals._term.Term:
        n = self.state.fix_point()

        # TODO(Ruben): this is a hack to handle the case where n is a square
        # it seems to work in practice but I have no clue why it works
        # it would be nice to have a proof that it works
        if not n:
            raise StopIteration()

        self.state.ingest(n)
        self.state.emit(n)
        return n


def sqrt(f: Union[Fraction, int]) -> reals._real.Real:
    p, q = f.as_integer_ratio()

    if p == 0:
        return reals._real.Real.from_int(0)

    if p < 0:
        raise ValueError()

    assert p > 0

    if q == 1:
        return reals._real.Real(SqrtComputation(p))

    # simply returning SqrtComputation(Homographic(0, p, q, 0)) works for most fractions except when p, q are squares
    # and q > 1 for q = 1 and p square a hack is in place which seems to work in practice, so this way of computing
    # works when p and q > 1 are squares

    # this is a variation on Real(SqrtComputation(p)) / Real(SqrtComputation(q)) with less overhead
    return reals._real.Real(
        reals._quadratic_computation.QuadraticComputation(
            SqrtComputation(p),
            SqrtComputation(q),
            (0, 1, 0, 0, 0, 0, 1, 0)
        ))



import reals._real
import reals._term
import reals._normalize

from typing import Generator


def e_term_generator() -> Generator[reals._term.Term, None, None]:
    yield 2
    k = 2
    while True:
        yield 1
        yield k
        yield 1
        k += 2


e = reals._real.Real(e_term_generator())


def pi_term_generator() -> Generator[reals._term.Term, None, None]:
    yield (0, 4)

    m = 1
    n = 1
    while True:
        yield (m, n)
        m += 2
        n += m


pi_unnormalized = reals._real.Real(pi_term_generator())

pi = reals._normalize.normalize(pi_unnormalized)


def phi_term_generator() -> Generator[reals._term.Term, None, None]:
    while True:
        yield 1


phi = reals._real.Real(phi_term_generator())



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from reals._real import Real

from enum import Enum, auto
from fractions import Fraction


EPSILON = Fraction(1, 100000)


class ComparisonResult(Enum):
    SMALLER = auto()
    GREATER = auto()
    UNKNOWN = auto()


# TODO(Ruben): this is very ugly, needs a rewrite
def compare(x: Real, y: Real, epsilon: Fraction = EPSILON) -> ComparisonResult:
    from reals.approximation import Approximation
    x_approximation = Approximation(x)
    x_approximation.improve()
    y_approximation = Approximation(y)
    y_approximation.improve()

    while True:
        x_lower = x_approximation.lower_bound_fraction()
        x_upper = x_approximation.upper_bound_fraction()
        y_lower = y_approximation.lower_bound_fraction()
        y_upper = y_approximation.upper_bound_fraction()

        if x_upper and y_lower and x_upper < y_lower:
            return ComparisonResult.SMALLER

        if x_lower and y_upper and y_upper < x_lower:
            return ComparisonResult.GREATER

        x_eps = x_approximation.epsilon_fraction()
        y_eps = y_approximation.epsilon_fraction()

        if not x_eps:
            x_approximation.improve()
        elif x_eps > epsilon / 2:
            x_approximation.improve()

        if not y_eps:
            y_approximation.improve()
        elif y_eps > epsilon / 2:
            y_approximation.improve()

        if x_eps is not None and x_eps < epsilon / 2 and y_eps is not None and y_eps < epsilon / 2:
            break

    return ComparisonResult.UNKNOWN



import reals._term
import reals._utils
import reals._computation
import reals._homographic

DEFAULT_MAX_INGESTIONS = 15


class AlgebraicComputation(reals._computation.Computation):
    def __init__(
            self,
            x: reals._computation.Computation,
            coeffs: tuple[int, int, int, int],
            max_ingestions=DEFAULT_MAX_INGESTIONS) -> None:
        a, b, c, d = coeffs
        self.state = reals._homographic.Homographic(a, b, c, d)
        self.x = x
        self.max_ingestions = max_ingestions
        self.terminated = False
        self.simple_mode = True

    def ingest_x(self) -> None:
        try:
            self.state.ingest(next(self.x))
        except StopIteration:
            self.terminated = self.state.ingest_inf()

    def __next__(self) -> reals._term.Term:
        if self.terminated:
            raise StopIteration()

        assert not (self.state.c == 0 and self.state.d == 0)

        ingestions = 0
        while True:
            self.simple_mode = self.simple_mode and ingestions <= self.max_ingestions

            if (self.state.c != 0 and
                    reals._utils.sign(self.state.c) == reals._utils.sign(self.state.c + self.state.d)):
                n1 = self.state.a // self.state.c
                n2 = (self.state.a + self.state.b) // (self.state.c + self.state.d)

                if n1 == n2:
                    self.simple_mode = True
                    self.terminated = self.state.emit(n1)
                    return n1
                elif not self.simple_mode and n1 == n2 + 1 or n1 + 1 == n2:
                    n = min(n1, n2)
                    m = max(n1, n2) - n + 1
                    assert m != 1
                    self.terminated = self.state.emit((n, m))
                    assert not self.terminated
                    return (n, m)

            self.ingest_x()
            if self.terminated:
                raise StopIteration()

            ingestions += 1



import reals._term
import reals._utils
import reals._computation
import reals._bihomographic


DEFAULT_MAX_INGESTIONS = 15


class QuadraticComputation(reals._computation.Computation):
    def __init__(self,
                 x: reals._computation.Computation,
                 y: reals._computation.Computation,
                 coeffs: tuple[int, int, int, int, int, int, int, int],
                 max_ingestions=DEFAULT_MAX_INGESTIONS) -> None:
        a, b, c, d, e, f, g, h = coeffs
        self.state = reals._bihomographic.Bihomographic(a, b, c, d, e, f, g, h)
        self.x, self.y = x, y
        self.max_ingestions = max_ingestions
        self.terminated = False
        self.simple_mode = True

    def ingest_x(self) -> None:
        try:
            self.state.x_ingest(next(self.x))
        except StopIteration:
            self.terminated = self.state.x_ingest_inf()

    def ingest_y(self) -> None:
        try:
            self.state.y_ingest(next(self.y))
        except StopIteration:
            self.terminated = self.state.y_ingest_inf()

    def close_enough(self, a: int, b: int) -> bool:
        return (a == b) or not self.simple_mode and (a == b + 1 or b == a + 1)

    def __next__(self) -> reals._term.Term:
        if self.terminated:
            raise StopIteration()

        assert not (self.state.e == 0 and self.state.f == 0 and
                    self.state.g == 0 and self.state.h == 0)

        ingestions = 0
        while True:
            self.simple_mode = self.simple_mode and ingestions < self.max_ingestions

            n00 = self.state.a + self.state.b + self.state.c + self.state.d
            d00 = self.state.e + self.state.f + self.state.g + self.state.h
            n10 = self.state.a + self.state.b
            d10 = self.state.e + self.state.f
            n01 = self.state.a + self.state.c
            d01 = self.state.e + self.state.g
            n11 = self.state.a
            d11 = self.state.e

            x_ingest = (d00 == 0 or d10 == 0 or d11 == 0 or
                        reals._utils.sign(d00) != reals._utils.sign(d10) or
                        reals._utils.sign(d01) != reals._utils.sign(d11))
            y_ingest = (d00 == 0 or d01 == 0 or d11 == 0 or
                        reals._utils.sign(d00) != reals._utils.sign(d01) or
                        reals._utils.sign(d10) != reals._utils.sign(d11))

            if not x_ingest and not y_ingest:
                q00 = n00 // d00
                q01 = n01 // d01
                q10 = n10 // d10
                q11 = n11 // d11
                too_far = not self.close_enough(q10, q01)
                x_ingest = too_far or not self.close_enough(q00, q10) or not self.close_enough(q01, q11)
                y_ingest = too_far or not self.close_enough(q00, q01) or not self.close_enough(q10, q11)

            if not x_ingest and not y_ingest:
                if self.simple_mode:
                    self.terminated = self.state.emit(q00)
                    return q00
                else:
                    n = min(q00, q10, q01, q11)
                    m = max(q00, q10, q01, q11) - n + 1
                    if m == 1:
                        self.simple_mode = True
                        self.terminated = self.state.emit(n)
                        return n
                    self.terminated = self.state.emit((n, m))
                    assert not self.terminated
                    return (n, m)

            ingestions += 1
            if x_ingest:
                self.ingest_x()
            if y_ingest:
                self.ingest_y()
            if self.terminated:
                raise StopIteration()



import reals._real
import reals._algebraic_computation


DEFAULT_MAX_INGESTIONS = 100


def normalize(x: reals._real.Real, max_ingestions=DEFAULT_MAX_INGESTIONS) -> reals._real.Real:
    return reals._real.Real(
        reals._algebraic_computation.AlgebraicComputation(x.compute(), (1, 0, 0, 1), max_ingestions))



import reals._term


class Bihomographic:
    a: int
    b: int
    c: int
    d: int
    e: int
    f: int
    g: int
    h: int

    def __init__(self, a: int, b: int, c: int, d: int, e: int, f: int, g: int, h: int) -> None:
        self.a, self.b, self.c, self.d, self.e, self.f, self.g, self.h = a, b, c, d, e, f, g, h

    # replace x by (n + m/x), and bring the resulting expression into biholomorphic form again
    def x_ingest(self, term: reals._term.Term) -> None:
        n, m = reals._term.expand_term(term)
        self.a, self.b, self.c, self.d = n * self.a + self.c, n * self.b + self.d, m * self.a, m * self.b
        self.e, self.f, self.g, self.h = n * self.e + self.g, n * self.f + self.h, m * self.e, m * self.f

    def x_ingest_inf(self) -> bool:
        self.c, self.d, = self.a, self.b
        self.g, self.h = self.e, self.f
        return self.e == 0 and self.f == 0 and self.g == 0 and self.h == 0

    # replace y by (n + m/y), and bring the resulting expression into biholomorphic form again
    def y_ingest(self, term: reals._term.Term) -> None:
        n, m = reals._term.expand_term(term)
        self.a, self.b, self.c, self.d = n * self.a + self.b, m * self.a, n * self.c + self.d, m * self.c
        self.e, self.f, self.g, self.h = n * self.e + self.f, m * self.e, n * self.g + self.h, m * self.g

    def y_ingest_inf(self) -> bool:
        self.b, self.d = self.a, self.c
        self.f, self.h = self.e, self.g
        return self.e == 0 and self.f == 0 and self.g == 0 and self.h == 0

    # replace the biholomorphic expression e by m / (e - n)
    def emit(self, term: reals._term.Term) -> bool:
        n, m = reals._term.expand_term(term)
        self.a, self.b, self.c, self.d = \
            self.a - n * self.e, self.b - n * self.f, self.c - n * self.g, self.d - n * self.h
        terminated = self.a == 0 and self.b == 0 and self.c == 0 and self.d == 0
        self.a, self.b, self.c, self.d, self.e, self.f, self.g, self.h = \
            m * self.e, m * self.f, m * self.g, m * self.h, self.a, self.b, self.c, self.d
        return terminated



from __future__ import annotations

import reals._term
import reals._inverse
import reals._computation
import reals._homographic
import reals._algebraic_computation
import reals._quadratic_computation

from decimal import Decimal
from fractions import Fraction
from typing import Generator, Iterable, Iterator, Union

DEFAULT_DIGITS = 5


class CachedComputation(reals._computation.Computation):
    def __init__(self, iterator: Iterator[reals._term.Term], cache: list[reals._term.Term]):
        self.cache = cache
        self.iterator = iterator
        self.index = 0

    def __next__(self) -> reals._term.Term:
        assert self.index <= len(self.cache)

        if self.index == len(self.cache):
            self.cache.append(next(self.iterator))

        index = self.index
        self.index += 1
        return self.cache[index]


class Real:
    def __init__(self, it: Union[Iterator[reals._term.Term], Iterable[reals._term.Term]]) -> None:
        self.iterator = iter(it)
        self.cache: list[reals._term.Term] = []

    def compute(self) -> reals._computation.Computation:
        return CachedComputation(self.iterator, self.cache)

    @staticmethod
    def from_number(x: Union[int, Fraction, Decimal]):
        if isinstance(x, int):
            return Real.from_int(x)

        p, q = x.as_integer_ratio()
        return Real(reals._algebraic_computation.AlgebraicComputation(iter([]), (p, p, q, q)))

    @staticmethod
    def from_int(n: int) -> 'Real':
        return Real([n])

    @staticmethod
    def from_fraction(f: Fraction) -> 'Real':
        p, q = f.as_integer_ratio()
        return Real(reals._algebraic_computation.AlgebraicComputation(iter([]), (p, p, q, q)))

    @staticmethod
    def from_decimal(d: Decimal) -> 'Real':
        p, q = d.as_integer_ratio()
        return Real(reals._algebraic_computation.AlgebraicComputation(iter([]), (p, p, q, q)))

    @staticmethod
    def from_float(f: float) -> 'Real':
        p, q = f.as_integer_ratio()
        return Real(reals._algebraic_computation.AlgebraicComputation(iter([]), (p, p, q, q)))

    @staticmethod
    def from_iter(i: Iterator[reals._term.Term]) -> 'Real':
        return Real(i)

    def inverse(self):
        return Real(reals._inverse.InverseComputation(self.compute()))

    def __neg__(self):
        return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (-1, 0, 0, 1)))

    def __mul__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (p, 0, 0, q)))
        elif isinstance(other, Real):
            return Real(reals._quadratic_computation.QuadraticComputation(
                    self.compute(),
                    other.compute(),
                    (1, 0, 0, 0, 0, 0, 0, 1)))
        else:
            raise TypeError()

    def __rmul__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (p, 0, 0, q)))
        else:
            raise TypeError()

    def __add__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (q, p, 0, q)))
        elif isinstance(other, Real):
            return Real(reals._quadratic_computation.QuadraticComputation(self.compute(), other.compute(),
                                                                          (0, 1, 1, 0, 0, 0, 0, 1)))
        else:
            raise TypeError()

    def __radd__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (q, p, 0, q)))
        else:
            raise TypeError()

    def __sub__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (q, -p, 0, q)))
        elif isinstance(other, Real):
            return Real(reals._quadratic_computation.QuadraticComputation(self.compute(),
                                                                          other.compute(), (0, 1, -1, 0, 0, 0, 0, 1)))
        else:
            raise TypeError()

    def __rsub__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (-q, p, 0, q)))
        else:
            raise TypeError()

    def __truediv__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (q, 0, 0, p)))
        elif isinstance(other, Real):
            return Real(reals._quadratic_computation.QuadraticComputation(self.compute(),
                                                                          other.compute(), (0, 1, 0, 0, 0, 0, 1, 0)))
        else:
            raise TypeError()

    def __rtruediv__(self, other):
        if isinstance(other, Fraction) or isinstance(other, int):
            p, q = other.as_integer_ratio()
            return Real(reals._algebraic_computation.AlgebraicComputation(self.compute(), (0, p, q, 0)))
        else:
            raise TypeError()

    def __lt__(self, other: Real) -> bool:
        from reals._compare import compare, ComparisonResult
        return compare(self, other) == ComparisonResult.SMALLER

    def __gt__(self, other: Real) -> bool:
        from reals._compare import compare, ComparisonResult
        return compare(self, other) == ComparisonResult.GREATER

    def __eq__(self, other) -> bool:
        from reals._compare import compare, ComparisonResult
        return compare(self, other) == ComparisonResult.UNKNOWN

    def evaluate(self, n: int) -> str:
        return rounded_digits(self, n)

    def to_decimal(self, n: int) -> Decimal:
        return Decimal(self.evaluate(n))

    def to_float(self) -> float:
        import reals.approximation  # this is ugly but necessary to avoid circular imports
        return reals.approximation.closest_float(self)  # noqa

    def __format__(self, spec):
        assert spec[0] == '.'
        assert spec[-1] == 'f'
        num_digits = int(spec[1:-1])
        return rounded_digits(self, num_digits)

    def __str__(self):
        return rounded_digits(self, DEFAULT_DIGITS)

    def __repr__(self):
        return (f'<{self.__class__.__module__}.{self.__class__.__name__} object at {hex(id(self))} '
                f'(approximate value: {str(self)})>')


def rounded_digits(x: Real, n: int) -> str:
    rounding_constant = Fraction(1, 2 * 10**n)
    return digits(x + rounding_constant, n)


def digits(x: Real, n: int) -> str:
    digit_generator = digits_helper(x.compute())
    digits = str(next(digit_generator))
    try:
        digits += '.'
        for _ in range(n):
            digits += next(digit_generator)
    except StopIteration:
        return digits
    return digits


def digits_helper(c: reals._computation.Computation) -> Generator[str, None, None]:
    h = reals._homographic.Homographic(1, 0, 0, 1)
    terminated = False
    is_negative = False

    assert not (h.c == 0 and h.d == 0)

    while not (h.c == 0 and h.d == 0):
        if h.c != 0 and h.c + h.d != 0:
            n1 = h.a // h.c
            n2 = (h.a + h.b) // (h.c + h.d)
            if n1 == n2:
                digit = n1

                h.a, h.b = h.a - digit * h.c, h.b - digit * h.d
                h.a, h.b = h.a * 10, h.b * 10

                if (not is_negative) and digit < 0:
                    is_negative = True
                    digit += 1
                    if digit == 0:
                        yield '-' + str(digit)
                    else:
                        yield str(digit)
                elif is_negative:
                    if digit > 0 and h.a == 0 and h.b == 0:
                        digit = 10 - digit
                        is_negative = False
                    else:
                        digit = 9 - digit
                    yield str(digit)
                else:
                    yield str(digit)

                continue

        assert not terminated
        try:
            h.ingest(next(c))
        except StopIteration:
            h.ingest_inf()
            terminated = True


Number = Union[int, Decimal, Fraction, Real]


def ensure_real(x: Number) -> Real:
    if isinstance(x, Real):
        return x
    if isinstance(x, Fraction):
        return Real.from_fraction(x)
    if isinstance(x, int):
        return Real.from_int(x)
    if isinstance(x, Decimal):
        return Real.from_decimal(x)



import reals._computation
import reals._term

from typing import Optional


class InverseComputation(reals._computation.Computation):
    def __init__(self, computation: reals._computation.Computation):
        self.computation = computation
        self.first = True
        self.cached_term: Optional[reals._term.Term] = None

    def __next__(self) -> reals._term.Term:
        if self.cached_term:
            term = self.cached_term
            self.cached_term = None
            return term

        next_term = next(self.computation)
        if self.first:
            self.first = False

            n, _ = reals._term.expand_term(next_term)

            if n == 0:
                return next(self.computation)
            else:
                self.cached_term = next_term
                return 0

        return next_term



import reals._term

from typing import Optional


class Homographic:
    a: int
    b: int
    c: int
    d: int

    def __init__(self, a: int, b: int, c: int, d: int) -> None:
        self.a, self.b, self.c, self.d = a, b, c, d

    # replace x by (n + m/x), and bring the resulting expression into homographic form again
    def ingest(self, term: reals._term.Term) -> None:
        n, m = reals._term.expand_term(term)
        self.a, self.b = n * self.a + self.b, m * self.a
        self.c, self.d = n * self.c + self.d, m * self.c

    def ingest_inf(self) -> bool:
        self.b, self.d = self.a, self.c
        return self.c == 0 and self.d == 0

    def emit(self, term: reals._term.Term) -> bool:
        n, m = reals._term.expand_term(term)
        self.a, self.b = self.a - n * self.c, self.b - n * self.d
        terminated = self.a == 0 and self.b == 0
        self.a, self.b, self.c, self.d = m * self.c, m * self.d, self.a, self.b
        return terminated

    # returns an integer n such that self.evaluate(n) = n or self.evaluate(n) = n + 1
    def fix_point(self) -> Optional[int]:
        current_guess = self.guess_int()

        if not current_guess:
            return None

        while True:
            value = self.evaluate_int(current_guess)
            assert value
            next_guess = (current_guess + value) // 2

            if next_guess == current_guess or next_guess == current_guess + 1:
                return current_guess

            current_guess = next_guess

    # evaluate the homographic form at an integer and floor the result
    # returns None if the denominator is zero when evaluated at n
    def evaluate_int(self, n: int) -> Optional[int]:
        denominator = self.c * n + self.d
        if denominator == 0:
            return None
        return (self.a * n + self.b) // denominator

    # returns an integer that makes the denominator positive, or None if it's impossible
    # (this happens when denominator is of form 0x + n for some nonpositive n)
    def guess_int(self) -> Optional[int]:
        if self.c == 0:
            if self.d <= 0:
                return None
            return 0

        if self.c > 0:
            return max(-self.d // self.c + 1, 0)
        else:
            assert self.c < 0
            return min(-(self.d // self.c) - 1, 0)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='reals',
    version='0.0.2',
    author='Ruben van Nieuwpoort',
    author_email='ruben@vannieuwpoort.dev',
    license='MIT',
    description='A lightweight python3 library for arithmetic with real numbers.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/rubenvannieuwpoort/reals',
    project_urls={
        'Bug Tracker': 'https://github.com/rubenvannieuwpoort/reals/issues',
    },
    packages=['reals'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)


import reals._real
import reals._constants
import reals._computation
import reals.approximation
import reals._exponential

from fractions import Fraction


def sin(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return _sin(x)


def sinh(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return (reals._exponential.exp(x) - reals._exponential.exp(-x)) / 2


def csc(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return 1 / _sin(x)


def csch(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return 2 / (reals._exponential.exp(x) - reals._exponential.exp(-x))


def cos(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return _cos(x)


def cosh(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return (reals._exponential.exp(x) + reals._exponential.exp(-x)) / 2


def sec(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return 1 / _cos(x)


def sech(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    return 2 / (reals._exponential.exp(x) + reals._exponential.exp(-x))


def tan(x: reals._real.Number) -> reals._real.Real:
    x = reals._real.ensure_real(x)
    return _sin(x) / _cos(x)


def tanh(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    exponential = reals._exponential.exp(2 * x)
    return (exponential - 1) / (exponential + 1)


def cot(x: reals._real.Number) -> reals._real.Real:
    x = reals._real.ensure_real(x)
    return _cos(x) / _sin(x)


def coth(x: reals._real.Number):
    x = reals._real.ensure_real(x)
    exponential = reals._exponential.exp(2 * x)
    return (exponential + 1) / (exponential - 1)


class SineComputation(reals._computation.Computation):
    def __init__(self, x: reals._real.Real) -> None:
        self.terms = 0
        self.n = 0
        self.current_term = x
        self.x2 = x * x
        self.hi = self.next_term()
        self.lo = self.hi + self.next_term()
        self.lo_comp = self.lo.compute()
        self.hi_comp = self.hi.compute()

    def next_term(self):
        term = self.current_term
        self.n += 2
        self.current_term = self.current_term * self.x2 / (self.n * (self.n + 1))
        return term

    def improve(self):
        self.hi = self.lo + self.next_term()
        self.lo = self.hi + self.next_term()
        self.lo_comp = self.lo.compute()
        self.hi_comp = self.hi.compute()
        for _ in range(0, self.terms):
            next(self.lo_comp)
            next(self.hi_comp)

    def __next__(self) -> tuple[int, int]:
        while True:
            n_lo = next(self.lo_comp)
            n_hi = next(self.hi_comp)

            if n_lo == n_hi:
                self.terms += 1
                return n_lo
            else:
                self.improve()


# set n=1 for cosine, n=2 for sin/x
class TrigComputation(reals._computation.Computation):
    def __init__(self, n: int, x: reals._real.Real) -> None:
        self.terms = 0
        self.n = n
        self.current_term = reals._real.Real.from_int(1)
        self.x2 = -x * x
        self.lo = self.next_term()
        self.hi = self.lo + self.next_term()
        self.lo_comp = self.lo.compute()
        self.hi_comp = self.hi.compute()

    def next_term(self):
        term = self.current_term
        self.current_term = self.current_term * self.x2 / (self.n * (self.n + 1))
        self.n += 2
        return term

    def improve(self):
        self.lo = self.hi + self.next_term()
        self.hi = self.lo + self.next_term()
        self.lo_comp = self.lo.compute()
        self.hi_comp = self.hi.compute()
        for _ in range(0, self.terms):
            next(self.lo_comp)
            next(self.hi_comp)

    def __next__(self) -> tuple[int, int]:
        while True:
            n_lo = next(self.lo_comp)
            n_hi = next(self.hi_comp)

            if n_lo == n_hi:
                self.terms += 1
                return n_lo
            else:
                self.improve()


# reduce to [-pi, pi] (plus or minus epsilon)
def reduce(x: reals._real.Real) -> reals._real.Real:
    a = reals.approximation.Approximation(x / (reals._constants.pi * 2) + Fraction(1, 2))
    epsilon = Fraction(1, 1000)
    a.improve_epsilon(epsilon)
    frac = a.as_fraction()
    assert frac
    p, q = frac.as_integer_ratio()
    return x - reals._constants.pi * 2 * (p // q)


def _sin(x: reals._real.Real) -> reals._real.Real:
    xr = reduce(x)
    a = reals.approximation.Approximation(2 * xr / reals._constants.pi + Fraction(1, 2))
    epsilon = Fraction(1, 1000)
    a.improve_epsilon(epsilon)
    frac = a.as_fraction()
    assert frac
    p, q = frac.as_integer_ratio()
    n = p // q
    if n == -2:
        return -_sin2(reals._constants.pi + xr)
    if n == -1:
        return -_cos2(xr + reals._constants.pi / 2)
    if n == 0:
        return _sin2(xr)
    if n == 1:
        return _cos2(xr - reals._constants.pi / 2)
    if n == 2:
        return _sin2(reals._constants.pi - xr)
    raise Exception()


def _cos(x: reals._real.Real) -> reals._real.Real:
    xr = reduce(x)
    a = reals.approximation.Approximation(2 * xr / reals._constants.pi + Fraction(1, 2))
    epsilon = Fraction(1, 1000)
    a.improve_epsilon(epsilon)
    frac = a.as_fraction()
    assert frac
    p, q = frac.as_integer_ratio()
    n = p // q
    if n == -2:
        return -_cos2(xr + reals._constants.pi)
    if n == -1:
        return _sin2(reals._constants.pi / 2 + xr)
    if n == 0:
        return _cos2(xr)
    if n == 1:
        return _sin2(reals._constants.pi / 2 - xr)
    if n == 2:
        return -_cos2(xr - reals._constants.pi)
    raise Exception()


def _sin2(x: reals._real.Real) -> reals._real.Real:
    return x * reals._real.Real(TrigComputation(2, x))


def _cos2(x: reals._real.Real) -> reals._real.Real:
    return reals._real.Real(TrigComputation(1, x))



from reals._real import Real
from reals._constants import e, phi, pi
from reals._exponential import exp
from reals._sqrt import sqrt
from reals._trigonometric import sin, sinh, csc, csch, cos, cosh, sec, sech, tan, tanh, cot, coth

__all__ = ['Real', 'e', 'phi', 'pi', 'exp', 'sqrt', 'sin', 'sinh', 'csc', 'csch', 'cos', 'cosh', 'sec', 'sech', 'tan',
           'tanh', 'cot', 'coth']
